#!/usr/bin/env python3

from typing import Tuple

import pandas as pd
from pandera import check_types
from pandera.typing import DataFrame

from app import schemas


@check_types
def create_plating_instructions(
    plating: DataFrame[schemas.ConstructWorksheetSchema],
    method: str,
    assemblyColumns: Tuple[str, str] = ("src_plate", "src_well"),
) -> DataFrame[schemas.PlatingInstructionsSchema]:
    """Generate instructions for plating

    Arguments
    ---------
    plating : pd.DataFrame
        construct worksheet dataframe used for
        actual assembly. This includes only
        assemblies that were actually attempted.

    method : str
        Type of plating. Can be either:
        8-wide or 6-wide or consecutive or qpix

    assemblyColumns : tuple of 2 strings, optional
        This option is used to denote alternative assembly
        plate names and wells in case the original assembly worksheet
        was modified. [default: ('src_plate', 'src_well')]

    Returns
    -------
    plating : pd.DataFrame
        Instructions/Map for how to plate
    """
    assert (
        (method == "8-wide")
        or (method == "6-wide")
        or (method == "consecutive")
        or (method == "qpix")
        or (method == "biomek")
    )
    plateColumn, wellColumn = assemblyColumns
    if method == "6-wide":
        platingMap = pd.DataFrame(
            {
                "96_WELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "A9",
                    9: "A10",
                    10: "A11",
                    11: "A12",
                    12: "B1",
                    13: "B2",
                    14: "B3",
                    15: "B4",
                    16: "B5",
                    17: "B6",
                    18: "B7",
                    19: "B8",
                    20: "B9",
                    21: "B10",
                    22: "B11",
                    23: "B12",
                    24: "C1",
                    25: "C2",
                    26: "C3",
                    27: "C4",
                    28: "C5",
                    29: "C6",
                    30: "C7",
                    31: "C8",
                    32: "C9",
                    33: "C10",
                    34: "C11",
                    35: "C12",
                    36: "D1",
                    37: "D2",
                    38: "D3",
                    39: "D4",
                    40: "D5",
                    41: "D6",
                    42: "D7",
                    43: "D8",
                    44: "D9",
                    45: "D10",
                    46: "D11",
                    47: "D12",
                    48: "E1",
                    49: "E2",
                    50: "E3",
                    51: "E4",
                    52: "E5",
                    53: "E6",
                    54: "E7",
                    55: "E8",
                    56: "E9",
                    57: "E10",
                    58: "E11",
                    59: "E12",
                    60: "F1",
                    61: "F2",
                    62: "F3",
                    63: "F4",
                    64: "F5",
                    65: "F6",
                    66: "F7",
                    67: "F8",
                    68: "F9",
                    69: "F10",
                    70: "F11",
                    71: "F12",
                    72: "G1",
                    73: "G2",
                    74: "G3",
                    75: "G4",
                    76: "G5",
                    77: "G6",
                    78: "G7",
                    79: "G8",
                    80: "G9",
                    81: "G10",
                    82: "G11",
                    83: "G12",
                    84: "H1",
                    85: "H2",
                    86: "H3",
                    87: "H4",
                    88: "H5",
                    89: "H6",
                    90: "H7",
                    91: "H8",
                    92: "H9",
                    93: "H10",
                    94: "H11",
                    95: "H12",
                },
                "QPLATE": {
                    0: "QPLATE_1",
                    1: "QPLATE_1",
                    2: "QPLATE_1",
                    3: "QPLATE_1",
                    4: "QPLATE_1",
                    5: "QPLATE_1",
                    6: "QPLATE_2",
                    7: "QPLATE_2",
                    8: "QPLATE_2",
                    9: "QPLATE_2",
                    10: "QPLATE_2",
                    11: "QPLATE_2",
                    12: "QPLATE_1",
                    13: "QPLATE_1",
                    14: "QPLATE_1",
                    15: "QPLATE_1",
                    16: "QPLATE_1",
                    17: "QPLATE_1",
                    18: "QPLATE_2",
                    19: "QPLATE_2",
                    20: "QPLATE_2",
                    21: "QPLATE_2",
                    22: "QPLATE_2",
                    23: "QPLATE_2",
                    24: "QPLATE_1",
                    25: "QPLATE_1",
                    26: "QPLATE_1",
                    27: "QPLATE_1",
                    28: "QPLATE_1",
                    29: "QPLATE_1",
                    30: "QPLATE_2",
                    31: "QPLATE_2",
                    32: "QPLATE_2",
                    33: "QPLATE_2",
                    34: "QPLATE_2",
                    35: "QPLATE_2",
                    36: "QPLATE_1",
                    37: "QPLATE_1",
                    38: "QPLATE_1",
                    39: "QPLATE_1",
                    40: "QPLATE_1",
                    41: "QPLATE_1",
                    42: "QPLATE_2",
                    43: "QPLATE_2",
                    44: "QPLATE_2",
                    45: "QPLATE_2",
                    46: "QPLATE_2",
                    47: "QPLATE_2",
                    48: "QPLATE_1",
                    49: "QPLATE_1",
                    50: "QPLATE_1",
                    51: "QPLATE_1",
                    52: "QPLATE_1",
                    53: "QPLATE_1",
                    54: "QPLATE_2",
                    55: "QPLATE_2",
                    56: "QPLATE_2",
                    57: "QPLATE_2",
                    58: "QPLATE_2",
                    59: "QPLATE_2",
                    60: "QPLATE_1",
                    61: "QPLATE_1",
                    62: "QPLATE_1",
                    63: "QPLATE_1",
                    64: "QPLATE_1",
                    65: "QPLATE_1",
                    66: "QPLATE_2",
                    67: "QPLATE_2",
                    68: "QPLATE_2",
                    69: "QPLATE_2",
                    70: "QPLATE_2",
                    71: "QPLATE_2",
                    72: "QPLATE_1",
                    73: "QPLATE_1",
                    74: "QPLATE_1",
                    75: "QPLATE_1",
                    76: "QPLATE_1",
                    77: "QPLATE_1",
                    78: "QPLATE_2",
                    79: "QPLATE_2",
                    80: "QPLATE_2",
                    81: "QPLATE_2",
                    82: "QPLATE_2",
                    83: "QPLATE_2",
                    84: "QPLATE_1",
                    85: "QPLATE_1",
                    86: "QPLATE_1",
                    87: "QPLATE_1",
                    88: "QPLATE_1",
                    89: "QPLATE_1",
                    90: "QPLATE_2",
                    91: "QPLATE_2",
                    92: "QPLATE_2",
                    93: "QPLATE_2",
                    94: "QPLATE_2",
                    95: "QPLATE_2",
                },
                "QWELL": {
                    0: "F1",
                    1: "E1",
                    2: "D1",
                    3: "C1",
                    4: "B1",
                    5: "A1",
                    6: "F1",
                    7: "E1",
                    8: "D1",
                    9: "C1",
                    10: "B1",
                    11: "A1",
                    12: "F2",
                    13: "E2",
                    14: "D2",
                    15: "C2",
                    16: "B2",
                    17: "A2",
                    18: "F2",
                    19: "E2",
                    20: "D2",
                    21: "C2",
                    22: "B2",
                    23: "A2",
                    24: "F3",
                    25: "E3",
                    26: "D3",
                    27: "C3",
                    28: "B3",
                    29: "A3",
                    30: "F3",
                    31: "E3",
                    32: "D3",
                    33: "C3",
                    34: "B3",
                    35: "A3",
                    36: "F4",
                    37: "E4",
                    38: "D4",
                    39: "C4",
                    40: "B4",
                    41: "A4",
                    42: "F4",
                    43: "E4",
                    44: "D4",
                    45: "C4",
                    46: "B4",
                    47: "A4",
                    48: "F5",
                    49: "E5",
                    50: "D5",
                    51: "C5",
                    52: "B5",
                    53: "A5",
                    54: "F5",
                    55: "E5",
                    56: "D5",
                    57: "C5",
                    58: "B5",
                    59: "A5",
                    60: "F6",
                    61: "E6",
                    62: "D6",
                    63: "C6",
                    64: "B6",
                    65: "A6",
                    66: "F6",
                    67: "E6",
                    68: "D6",
                    69: "C6",
                    70: "B6",
                    71: "A6",
                    72: "F7",
                    73: "E7",
                    74: "D7",
                    75: "C7",
                    76: "B7",
                    77: "A7",
                    78: "F7",
                    79: "E7",
                    80: "D7",
                    81: "C7",
                    82: "B7",
                    83: "A7",
                    84: "F8",
                    85: "E8",
                    86: "D8",
                    87: "C8",
                    88: "B8",
                    89: "A8",
                    90: "F8",
                    91: "E8",
                    92: "D8",
                    93: "C8",
                    94: "B8",
                    95: "A8",
                },
            }
        )
    elif method == "8-wide":
        platingMap = pd.DataFrame(
            {
                "96_WELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "A9",
                    9: "A10",
                    10: "A11",
                    11: "A12",
                    12: "B1",
                    13: "B2",
                    14: "B3",
                    15: "B4",
                    16: "B5",
                    17: "B6",
                    18: "B7",
                    19: "B8",
                    20: "B9",
                    21: "B10",
                    22: "B11",
                    23: "B12",
                    24: "C1",
                    25: "C2",
                    26: "C3",
                    27: "C4",
                    28: "C5",
                    29: "C6",
                    30: "C7",
                    31: "C8",
                    32: "C9",
                    33: "C10",
                    34: "C11",
                    35: "C12",
                    36: "D1",
                    37: "D2",
                    38: "D3",
                    39: "D4",
                    40: "D5",
                    41: "D6",
                    42: "D7",
                    43: "D8",
                    44: "D9",
                    45: "D10",
                    46: "D11",
                    47: "D12",
                    48: "E1",
                    49: "E2",
                    50: "E3",
                    51: "E4",
                    52: "E5",
                    53: "E6",
                    54: "E7",
                    55: "E8",
                    56: "E9",
                    57: "E10",
                    58: "E11",
                    59: "E12",
                    60: "F1",
                    61: "F2",
                    62: "F3",
                    63: "F4",
                    64: "F5",
                    65: "F6",
                    66: "F7",
                    67: "F8",
                    68: "F9",
                    69: "F10",
                    70: "F11",
                    71: "F12",
                    72: "G1",
                    73: "G2",
                    74: "G3",
                    75: "G4",
                    76: "G5",
                    77: "G6",
                    78: "G7",
                    79: "G8",
                    80: "G9",
                    81: "G10",
                    82: "G11",
                    83: "G12",
                    84: "H1",
                    85: "H2",
                    86: "H3",
                    87: "H4",
                    88: "H5",
                    89: "H6",
                    90: "H7",
                    91: "H8",
                    92: "H9",
                    93: "H10",
                    94: "H11",
                    95: "H12",
                },
                "QPLATE": {
                    0: "QPLATE_1",
                    1: "QPLATE_1",
                    2: "QPLATE_1",
                    3: "QPLATE_1",
                    4: "QPLATE_1",
                    5: "QPLATE_1",
                    6: "QPLATE_1",
                    7: "QPLATE_1",
                    8: "QPLATE_2",
                    9: "QPLATE_2",
                    10: "QPLATE_2",
                    11: "QPLATE_2",
                    12: "QPLATE_1",
                    13: "QPLATE_1",
                    14: "QPLATE_1",
                    15: "QPLATE_1",
                    16: "QPLATE_1",
                    17: "QPLATE_1",
                    18: "QPLATE_1",
                    19: "QPLATE_1",
                    20: "QPLATE_2",
                    21: "QPLATE_2",
                    22: "QPLATE_2",
                    23: "QPLATE_2",
                    24: "QPLATE_1",
                    25: "QPLATE_1",
                    26: "QPLATE_1",
                    27: "QPLATE_1",
                    28: "QPLATE_1",
                    29: "QPLATE_1",
                    30: "QPLATE_1",
                    31: "QPLATE_1",
                    32: "QPLATE_2",
                    33: "QPLATE_2",
                    34: "QPLATE_2",
                    35: "QPLATE_2",
                    36: "QPLATE_1",
                    37: "QPLATE_1",
                    38: "QPLATE_1",
                    39: "QPLATE_1",
                    40: "QPLATE_1",
                    41: "QPLATE_1",
                    42: "QPLATE_1",
                    43: "QPLATE_1",
                    44: "QPLATE_2",
                    45: "QPLATE_2",
                    46: "QPLATE_2",
                    47: "QPLATE_2",
                    48: "QPLATE_1",
                    49: "QPLATE_1",
                    50: "QPLATE_1",
                    51: "QPLATE_1",
                    52: "QPLATE_1",
                    53: "QPLATE_1",
                    54: "QPLATE_1",
                    55: "QPLATE_1",
                    56: "QPLATE_2",
                    57: "QPLATE_2",
                    58: "QPLATE_2",
                    59: "QPLATE_2",
                    60: "QPLATE_1",
                    61: "QPLATE_1",
                    62: "QPLATE_1",
                    63: "QPLATE_1",
                    64: "QPLATE_1",
                    65: "QPLATE_1",
                    66: "QPLATE_1",
                    67: "QPLATE_1",
                    68: "QPLATE_2",
                    69: "QPLATE_2",
                    70: "QPLATE_2",
                    71: "QPLATE_2",
                    72: "QPLATE_2",
                    73: "QPLATE_2",
                    74: "QPLATE_2",
                    75: "QPLATE_2",
                    76: "QPLATE_2",
                    77: "QPLATE_2",
                    78: "QPLATE_2",
                    79: "QPLATE_2",
                    80: "QPLATE_2",
                    81: "QPLATE_2",
                    82: "QPLATE_2",
                    83: "QPLATE_2",
                    84: "QPLATE_2",
                    85: "QPLATE_2",
                    86: "QPLATE_2",
                    87: "QPLATE_2",
                    88: "QPLATE_2",
                    89: "QPLATE_2",
                    90: "QPLATE_2",
                    91: "QPLATE_2",
                    92: "QPLATE_2",
                    93: "QPLATE_2",
                    94: "QPLATE_2",
                    95: "QPLATE_2",
                },
                "QWELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "A1",
                    9: "A2",
                    10: "A3",
                    11: "A4",
                    12: "B1",
                    13: "B2",
                    14: "B3",
                    15: "B4",
                    16: "B5",
                    17: "B6",
                    18: "B7",
                    19: "B8",
                    20: "B1",
                    21: "B2",
                    22: "B3",
                    23: "B4",
                    24: "C1",
                    25: "C2",
                    26: "C3",
                    27: "C4",
                    28: "C5",
                    29: "C6",
                    30: "C7",
                    31: "C8",
                    32: "C1",
                    33: "C2",
                    34: "C3",
                    35: "C4",
                    36: "D1",
                    37: "D2",
                    38: "D3",
                    39: "D4",
                    40: "D5",
                    41: "D6",
                    42: "D7",
                    43: "D8",
                    44: "D1",
                    45: "D2",
                    46: "D3",
                    47: "D4",
                    48: "E1",
                    49: "E2",
                    50: "E3",
                    51: "E4",
                    52: "E5",
                    53: "E6",
                    54: "E7",
                    55: "E8",
                    56: "E1",
                    57: "E2",
                    58: "E3",
                    59: "E4",
                    60: "F1",
                    61: "F2",
                    62: "F3",
                    63: "F4",
                    64: "F5",
                    65: "F6",
                    66: "F7",
                    67: "F8",
                    68: "F1",
                    69: "F2",
                    70: "F3",
                    71: "F4",
                    72: "A5",
                    73: "A6",
                    74: "A7",
                    75: "A8",
                    76: "B5",
                    77: "B6",
                    78: "B7",
                    79: "B8",
                    80: "C5",
                    81: "C6",
                    82: "C7",
                    83: "C8",
                    84: "D5",
                    85: "D6",
                    86: "D7",
                    87: "D8",
                    88: "E5",
                    89: "E6",
                    90: "E7",
                    91: "E8",
                    92: "F5",
                    93: "F6",
                    94: "F7",
                    95: "F8",
                },
            }
        )
    elif method == "consecutive":
        platingMap = pd.DataFrame(
            {
                "96_WELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "A9",
                    9: "A10",
                    10: "A11",
                    11: "A12",
                    12: "B1",
                    13: "B2",
                    14: "B3",
                    15: "B4",
                    16: "B5",
                    17: "B6",
                    18: "B7",
                    19: "B8",
                    20: "B9",
                    21: "B10",
                    22: "B11",
                    23: "B12",
                    24: "C1",
                    25: "C2",
                    26: "C3",
                    27: "C4",
                    28: "C5",
                    29: "C6",
                    30: "C7",
                    31: "C8",
                    32: "C9",
                    33: "C10",
                    34: "C11",
                    35: "C12",
                    36: "D1",
                    37: "D2",
                    38: "D3",
                    39: "D4",
                    40: "D5",
                    41: "D6",
                    42: "D7",
                    43: "D8",
                    44: "D9",
                    45: "D10",
                    46: "D11",
                    47: "D12",
                    48: "E1",
                    49: "E2",
                    50: "E3",
                    51: "E4",
                    52: "E5",
                    53: "E6",
                    54: "E7",
                    55: "E8",
                    56: "E9",
                    57: "E10",
                    58: "E11",
                    59: "E12",
                    60: "F1",
                    61: "F2",
                    62: "F3",
                    63: "F4",
                    64: "F5",
                    65: "F6",
                    66: "F7",
                    67: "F8",
                    68: "F9",
                    69: "F10",
                    70: "F11",
                    71: "F12",
                    72: "G1",
                    73: "G2",
                    74: "G3",
                    75: "G4",
                    76: "G5",
                    77: "G6",
                    78: "G7",
                    79: "G8",
                    80: "G9",
                    81: "G10",
                    82: "G11",
                    83: "G12",
                    84: "H1",
                    85: "H2",
                    86: "H3",
                    87: "H4",
                    88: "H5",
                    89: "H6",
                    90: "H7",
                    91: "H8",
                    92: "H9",
                    93: "H10",
                    94: "H11",
                    95: "H12",
                },
                "QPLATE": {
                    0: "QPLATE_1",
                    1: "QPLATE_1",
                    2: "QPLATE_1",
                    3: "QPLATE_1",
                    4: "QPLATE_1",
                    5: "QPLATE_1",
                    6: "QPLATE_1",
                    7: "QPLATE_1",
                    8: "QPLATE_1",
                    9: "QPLATE_1",
                    10: "QPLATE_1",
                    11: "QPLATE_1",
                    12: "QPLATE_1",
                    13: "QPLATE_1",
                    14: "QPLATE_1",
                    15: "QPLATE_1",
                    16: "QPLATE_1",
                    17: "QPLATE_1",
                    18: "QPLATE_1",
                    19: "QPLATE_1",
                    20: "QPLATE_1",
                    21: "QPLATE_1",
                    22: "QPLATE_1",
                    23: "QPLATE_1",
                    24: "QPLATE_1",
                    25: "QPLATE_1",
                    26: "QPLATE_1",
                    27: "QPLATE_1",
                    28: "QPLATE_1",
                    29: "QPLATE_1",
                    30: "QPLATE_1",
                    31: "QPLATE_1",
                    32: "QPLATE_1",
                    33: "QPLATE_1",
                    34: "QPLATE_1",
                    35: "QPLATE_1",
                    36: "QPLATE_1",
                    37: "QPLATE_1",
                    38: "QPLATE_1",
                    39: "QPLATE_1",
                    40: "QPLATE_1",
                    41: "QPLATE_1",
                    42: "QPLATE_1",
                    43: "QPLATE_1",
                    44: "QPLATE_1",
                    45: "QPLATE_1",
                    46: "QPLATE_1",
                    47: "QPLATE_1",
                    48: "QPLATE_2",
                    49: "QPLATE_2",
                    50: "QPLATE_2",
                    51: "QPLATE_2",
                    52: "QPLATE_2",
                    53: "QPLATE_2",
                    54: "QPLATE_2",
                    55: "QPLATE_2",
                    56: "QPLATE_2",
                    57: "QPLATE_2",
                    58: "QPLATE_2",
                    59: "QPLATE_2",
                    60: "QPLATE_2",
                    61: "QPLATE_2",
                    62: "QPLATE_2",
                    63: "QPLATE_2",
                    64: "QPLATE_2",
                    65: "QPLATE_2",
                    66: "QPLATE_2",
                    67: "QPLATE_2",
                    68: "QPLATE_2",
                    69: "QPLATE_2",
                    70: "QPLATE_2",
                    71: "QPLATE_2",
                    72: "QPLATE_2",
                    73: "QPLATE_2",
                    74: "QPLATE_2",
                    75: "QPLATE_2",
                    76: "QPLATE_2",
                    77: "QPLATE_2",
                    78: "QPLATE_2",
                    79: "QPLATE_2",
                    80: "QPLATE_2",
                    81: "QPLATE_2",
                    82: "QPLATE_2",
                    83: "QPLATE_2",
                    84: "QPLATE_2",
                    85: "QPLATE_2",
                    86: "QPLATE_2",
                    87: "QPLATE_2",
                    88: "QPLATE_2",
                    89: "QPLATE_2",
                    90: "QPLATE_2",
                    91: "QPLATE_2",
                    92: "QPLATE_2",
                    93: "QPLATE_2",
                    94: "QPLATE_2",
                    95: "QPLATE_2",
                },
                "QWELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "B1",
                    9: "B2",
                    10: "B3",
                    11: "B4",
                    12: "B5",
                    13: "B6",
                    14: "B7",
                    15: "B8",
                    16: "C1",
                    17: "C2",
                    18: "C3",
                    19: "C4",
                    20: "C5",
                    21: "C6",
                    22: "C7",
                    23: "C8",
                    24: "D1",
                    25: "D2",
                    26: "D3",
                    27: "D4",
                    28: "D5",
                    29: "D6",
                    30: "D7",
                    31: "D8",
                    32: "E1",
                    33: "E2",
                    34: "E3",
                    35: "E4",
                    36: "E5",
                    37: "E6",
                    38: "E7",
                    39: "E8",
                    40: "F1",
                    41: "F2",
                    42: "F3",
                    43: "F4",
                    44: "F5",
                    45: "F6",
                    46: "F7",
                    47: "F8",
                    48: "A1",
                    49: "A2",
                    50: "A3",
                    51: "A4",
                    52: "A5",
                    53: "A6",
                    54: "A7",
                    55: "A8",
                    56: "B1",
                    57: "B2",
                    58: "B3",
                    59: "B4",
                    60: "B5",
                    61: "B6",
                    62: "B7",
                    63: "B8",
                    64: "C1",
                    65: "C2",
                    66: "C3",
                    67: "C4",
                    68: "C5",
                    69: "C6",
                    70: "C7",
                    71: "C8",
                    72: "D1",
                    73: "D2",
                    74: "D3",
                    75: "D4",
                    76: "D5",
                    77: "D6",
                    78: "D7",
                    79: "D8",
                    80: "E1",
                    81: "E2",
                    82: "E3",
                    83: "E4",
                    84: "E5",
                    85: "E6",
                    86: "E7",
                    87: "E8",
                    88: "F1",
                    89: "F2",
                    90: "F3",
                    91: "F4",
                    92: "F5",
                    93: "F6",
                    94: "F7",
                    95: "F8",
                },
            }
        )
    elif method == "qpix":
        platingMap = pd.DataFrame(
            {
                "96_WELL": {
                    0: "A1",
                    1: "B1",
                    2: "C1",
                    3: "D1",
                    4: "E1",
                    5: "F1",
                    6: "G1",
                    7: "H1",
                    8: "A2",
                    9: "B2",
                    10: "C2",
                    11: "D2",
                    12: "E2",
                    13: "F2",
                    14: "G2",
                    15: "H2",
                    16: "A3",
                    17: "B3",
                    18: "C3",
                    19: "D3",
                    20: "E3",
                    21: "F3",
                    22: "G3",
                    23: "H3",
                    24: "A4",
                    25: "B4",
                    26: "C4",
                    27: "D4",
                    28: "E4",
                    29: "F4",
                    30: "G4",
                    31: "H4",
                    32: "A5",
                    33: "B5",
                    34: "C5",
                    35: "D5",
                    36: "E5",
                    37: "F5",
                    38: "G5",
                    39: "H5",
                    40: "A6",
                    41: "B6",
                    42: "C6",
                    43: "D6",
                    44: "E6",
                    45: "F6",
                    46: "G6",
                    47: "H6",
                    48: "A7",
                    49: "B7",
                    50: "C7",
                    51: "D7",
                    52: "E7",
                    53: "F7",
                    54: "G7",
                    55: "H7",
                    56: "A8",
                    57: "B8",
                    58: "C8",
                    59: "D8",
                    60: "E8",
                    61: "F8",
                    62: "G8",
                    63: "H8",
                    64: "A9",
                    65: "B9",
                    66: "C9",
                    67: "D9",
                    68: "E9",
                    69: "F9",
                    70: "G9",
                    71: "H9",
                    72: "A10",
                    73: "B10",
                    74: "C10",
                    75: "D10",
                    76: "E10",
                    77: "F10",
                    78: "G10",
                    79: "H10",
                    80: "A11",
                    81: "B11",
                    82: "C11",
                    83: "D11",
                    84: "E11",
                    85: "F11",
                    86: "G11",
                    87: "H11",
                    88: "A12",
                    89: "B12",
                    90: "C12",
                    91: "D12",
                    92: "E12",
                    93: "F12",
                    94: "G12",
                    95: "H12",
                },
                "QPLATE": {
                    0: "QPLATE_1",
                    1: "QPLATE_1",
                    2: "QPLATE_1",
                    3: "QPLATE_1",
                    4: "QPLATE_1",
                    5: "QPLATE_1",
                    6: "QPLATE_1",
                    7: "QPLATE_1",
                    8: "QPLATE_1",
                    9: "QPLATE_1",
                    10: "QPLATE_1",
                    11: "QPLATE_1",
                    12: "QPLATE_1",
                    13: "QPLATE_1",
                    14: "QPLATE_1",
                    15: "QPLATE_1",
                    16: "QPLATE_1",
                    17: "QPLATE_1",
                    18: "QPLATE_1",
                    19: "QPLATE_1",
                    20: "QPLATE_1",
                    21: "QPLATE_1",
                    22: "QPLATE_1",
                    23: "QPLATE_1",
                    24: "QPLATE_1",
                    25: "QPLATE_1",
                    26: "QPLATE_1",
                    27: "QPLATE_1",
                    28: "QPLATE_1",
                    29: "QPLATE_1",
                    30: "QPLATE_1",
                    31: "QPLATE_1",
                    32: "QPLATE_1",
                    33: "QPLATE_1",
                    34: "QPLATE_1",
                    35: "QPLATE_1",
                    36: "QPLATE_1",
                    37: "QPLATE_1",
                    38: "QPLATE_1",
                    39: "QPLATE_1",
                    40: "QPLATE_1",
                    41: "QPLATE_1",
                    42: "QPLATE_1",
                    43: "QPLATE_1",
                    44: "QPLATE_1",
                    45: "QPLATE_1",
                    46: "QPLATE_1",
                    47: "QPLATE_1",
                    48: "QPLATE_2",
                    49: "QPLATE_2",
                    50: "QPLATE_2",
                    51: "QPLATE_2",
                    52: "QPLATE_2",
                    53: "QPLATE_2",
                    54: "QPLATE_2",
                    55: "QPLATE_2",
                    56: "QPLATE_2",
                    57: "QPLATE_2",
                    58: "QPLATE_2",
                    59: "QPLATE_2",
                    60: "QPLATE_2",
                    61: "QPLATE_2",
                    62: "QPLATE_2",
                    63: "QPLATE_2",
                    64: "QPLATE_2",
                    65: "QPLATE_2",
                    66: "QPLATE_2",
                    67: "QPLATE_2",
                    68: "QPLATE_2",
                    69: "QPLATE_2",
                    70: "QPLATE_2",
                    71: "QPLATE_2",
                    72: "QPLATE_2",
                    73: "QPLATE_2",
                    74: "QPLATE_2",
                    75: "QPLATE_2",
                    76: "QPLATE_2",
                    77: "QPLATE_2",
                    78: "QPLATE_2",
                    79: "QPLATE_2",
                    80: "QPLATE_2",
                    81: "QPLATE_2",
                    82: "QPLATE_2",
                    83: "QPLATE_2",
                    84: "QPLATE_2",
                    85: "QPLATE_2",
                    86: "QPLATE_2",
                    87: "QPLATE_2",
                    88: "QPLATE_2",
                    89: "QPLATE_2",
                    90: "QPLATE_2",
                    91: "QPLATE_2",
                    92: "QPLATE_2",
                    93: "QPLATE_2",
                    94: "QPLATE_2",
                    95: "QPLATE_2",
                },
                "QWELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "B1",
                    5: "B2",
                    6: "B3",
                    7: "B4",
                    8: "C1",
                    9: "C2",
                    10: "C3",
                    11: "C4",
                    12: "D1",
                    13: "D2",
                    14: "D3",
                    15: "D4",
                    16: "E1",
                    17: "E2",
                    18: "E3",
                    19: "E4",
                    20: "F1",
                    21: "F2",
                    22: "F3",
                    23: "F4",
                    24: "A5",
                    25: "A6",
                    26: "A7",
                    27: "A8",
                    28: "B5",
                    29: "B6",
                    30: "B7",
                    31: "B8",
                    32: "C5",
                    33: "C6",
                    34: "C7",
                    35: "C8",
                    36: "D5",
                    37: "D6",
                    38: "D7",
                    39: "D8",
                    40: "E5",
                    41: "E6",
                    42: "E7",
                    43: "E8",
                    44: "F5",
                    45: "F6",
                    46: "F7",
                    47: "F8",
                    48: "A1",
                    49: "A2",
                    50: "A3",
                    51: "A4",
                    52: "B1",
                    53: "B2",
                    54: "B3",
                    55: "B4",
                    56: "C1",
                    57: "C2",
                    58: "C3",
                    59: "C4",
                    60: "D1",
                    61: "D2",
                    62: "D3",
                    63: "D4",
                    64: "E1",
                    65: "E2",
                    66: "E3",
                    67: "E4",
                    68: "F1",
                    69: "F2",
                    70: "F3",
                    71: "F4",
                    72: "A5",
                    73: "A6",
                    74: "A7",
                    75: "A8",
                    76: "B5",
                    77: "B6",
                    78: "B7",
                    79: "B8",
                    80: "C5",
                    81: "C6",
                    82: "C7",
                    83: "C8",
                    84: "D5",
                    85: "D6",
                    86: "D7",
                    87: "D8",
                    88: "E5",
                    89: "E6",
                    90: "E7",
                    91: "E8",
                    92: "F5",
                    93: "F6",
                    94: "F7",
                    95: "F8",
                },
            }
        )
    elif method == "biomek":
        platingMap = pd.DataFrame(
            {
                "96_WELL": {
                    0: "A1",
                    1: "B1",
                    2: "C1",
                    3: "D1",
                    4: "E1",
                    5: "F1",
                    6: "G1",
                    7: "H1",
                    8: "A2",
                    9: "B2",
                    10: "C2",
                    11: "D2",
                    12: "E2",
                    13: "F2",
                    14: "G2",
                    15: "H2",
                    16: "A3",
                    17: "B3",
                    18: "C3",
                    19: "D3",
                    20: "E3",
                    21: "F3",
                    22: "G3",
                    23: "H3",
                    24: "A4",
                    25: "B4",
                    26: "C4",
                    27: "D4",
                    28: "E4",
                    29: "F4",
                    30: "G4",
                    31: "H4",
                    32: "A5",
                    33: "B5",
                    34: "C5",
                    35: "D5",
                    36: "E5",
                    37: "F5",
                    38: "G5",
                    39: "H5",
                    40: "A6",
                    41: "B6",
                    42: "C6",
                    43: "D6",
                    44: "E6",
                    45: "F6",
                    46: "G6",
                    47: "H6",
                    48: "A7",
                    49: "B7",
                    50: "C7",
                    51: "D7",
                    52: "E7",
                    53: "F7",
                    54: "G7",
                    55: "H7",
                    56: "A8",
                    57: "B8",
                    58: "C8",
                    59: "D8",
                    60: "E8",
                    61: "F8",
                    62: "G8",
                    63: "H8",
                    64: "A9",
                    65: "B9",
                    66: "C9",
                    67: "D9",
                    68: "E9",
                    69: "F9",
                    70: "G9",
                    71: "H9",
                    72: "A10",
                    73: "B10",
                    74: "C10",
                    75: "D10",
                    76: "E10",
                    77: "F10",
                    78: "G10",
                    79: "H10",
                    80: "A11",
                    81: "B11",
                    82: "C11",
                    83: "D11",
                    84: "E11",
                    85: "F11",
                    86: "G11",
                    87: "H11",
                    88: "A12",
                    89: "B12",
                    90: "C12",
                    91: "D12",
                    92: "E12",
                    93: "F12",
                    94: "G12",
                    95: "H12",
                },
                "QPLATE": {
                    0: "QPLATE_1",
                    1: "QPLATE_1",
                    2: "QPLATE_1",
                    3: "QPLATE_1",
                    4: "QPLATE_1",
                    5: "QPLATE_1",
                    6: "QPLATE_1",
                    7: "QPLATE_1",
                    8: "QPLATE_1",
                    9: "QPLATE_1",
                    10: "QPLATE_1",
                    11: "QPLATE_1",
                    12: "QPLATE_1",
                    13: "QPLATE_1",
                    14: "QPLATE_1",
                    15: "QPLATE_1",
                    16: "QPLATE_1",
                    17: "QPLATE_1",
                    18: "QPLATE_1",
                    19: "QPLATE_1",
                    20: "QPLATE_1",
                    21: "QPLATE_1",
                    22: "QPLATE_1",
                    23: "QPLATE_1",
                    24: "QPLATE_1",
                    25: "QPLATE_1",
                    26: "QPLATE_1",
                    27: "QPLATE_1",
                    28: "QPLATE_1",
                    29: "QPLATE_1",
                    30: "QPLATE_1",
                    31: "QPLATE_1",
                    32: "QPLATE_1",
                    33: "QPLATE_1",
                    34: "QPLATE_1",
                    35: "QPLATE_1",
                    36: "QPLATE_1",
                    37: "QPLATE_1",
                    38: "QPLATE_1",
                    39: "QPLATE_1",
                    40: "QPLATE_1",
                    41: "QPLATE_1",
                    42: "QPLATE_1",
                    43: "QPLATE_1",
                    44: "QPLATE_1",
                    45: "QPLATE_1",
                    46: "QPLATE_1",
                    47: "QPLATE_1",
                    48: "QPLATE_2",
                    49: "QPLATE_2",
                    50: "QPLATE_2",
                    51: "QPLATE_2",
                    52: "QPLATE_2",
                    53: "QPLATE_2",
                    54: "QPLATE_2",
                    55: "QPLATE_2",
                    56: "QPLATE_2",
                    57: "QPLATE_2",
                    58: "QPLATE_2",
                    59: "QPLATE_2",
                    60: "QPLATE_2",
                    61: "QPLATE_2",
                    62: "QPLATE_2",
                    63: "QPLATE_2",
                    64: "QPLATE_2",
                    65: "QPLATE_2",
                    66: "QPLATE_2",
                    67: "QPLATE_2",
                    68: "QPLATE_2",
                    69: "QPLATE_2",
                    70: "QPLATE_2",
                    71: "QPLATE_2",
                    72: "QPLATE_2",
                    73: "QPLATE_2",
                    74: "QPLATE_2",
                    75: "QPLATE_2",
                    76: "QPLATE_2",
                    77: "QPLATE_2",
                    78: "QPLATE_2",
                    79: "QPLATE_2",
                    80: "QPLATE_2",
                    81: "QPLATE_2",
                    82: "QPLATE_2",
                    83: "QPLATE_2",
                    84: "QPLATE_2",
                    85: "QPLATE_2",
                    86: "QPLATE_2",
                    87: "QPLATE_2",
                    88: "QPLATE_2",
                    89: "QPLATE_2",
                    90: "QPLATE_2",
                    91: "QPLATE_2",
                    92: "QPLATE_2",
                    93: "QPLATE_2",
                    94: "QPLATE_2",
                    95: "QPLATE_2",
                },
                "QWELL": {
                    0: "A1",
                    1: "A2",
                    2: "A3",
                    3: "A4",
                    4: "A5",
                    5: "A6",
                    6: "A7",
                    7: "A8",
                    8: "B1",
                    9: "B2",
                    10: "B3",
                    11: "B4",
                    12: "B5",
                    13: "B6",
                    14: "B7",
                    15: "B8",
                    16: "C1",
                    17: "C2",
                    18: "C3",
                    19: "C4",
                    20: "C5",
                    21: "C6",
                    22: "C7",
                    23: "C8",
                    24: "D1",
                    25: "D2",
                    26: "D3",
                    27: "D4",
                    28: "D5",
                    29: "D6",
                    30: "D7",
                    31: "D8",
                    32: "E1",
                    33: "E2",
                    34: "E3",
                    35: "E4",
                    36: "E5",
                    37: "E6",
                    38: "E7",
                    39: "E8",
                    40: "F1",
                    41: "F2",
                    42: "F3",
                    43: "F4",
                    44: "F5",
                    45: "F6",
                    46: "F7",
                    47: "F8",
                    48: "A1",
                    49: "A2",
                    50: "A3",
                    51: "A4",
                    52: "A5",
                    53: "A6",
                    54: "A7",
                    55: "A8",
                    56: "B1",
                    57: "B2",
                    58: "B3",
                    59: "B4",
                    60: "B5",
                    61: "B6",
                    62: "B7",
                    63: "B8",
                    64: "C1",
                    65: "C2",
                    66: "C3",
                    67: "C4",
                    68: "C5",
                    69: "C6",
                    70: "C7",
                    71: "C8",
                    72: "D1",
                    73: "D2",
                    74: "D3",
                    75: "D4",
                    76: "D5",
                    77: "D6",
                    78: "D7",
                    79: "D8",
                    80: "E1",
                    81: "E2",
                    82: "E3",
                    83: "E4",
                    84: "E5",
                    85: "E6",
                    86: "E7",
                    87: "E8",
                    88: "F1",
                    89: "F2",
                    90: "F3",
                    91: "F4",
                    92: "F5",
                    93: "F6",
                    94: "F7",
                    95: "F8",
                },
            }
        )
    instructions = plating.copy()
    instructions["QPLATE"] = plating.apply(
        lambda row: row[plateColumn]
        + "_"
        + platingMap.loc[platingMap["96_WELL"] == row[wellColumn], "QPLATE"].values[0],
        axis=1,
    )
    instructions["QWELL"] = plating.apply(
        lambda row: platingMap.loc[
            platingMap["96_WELL"] == row[wellColumn], "QWELL"
        ].values[0],
        axis=1,
    )
    return instructions
